/*
 * Copyright (c) 2001, DecisionSoft Limited All rights reserved.
 * Please see LICENSE.TXT for more information.
 */

//System includes

#ifdef WIN32
#include <iostream.h>
#endif

//Xerces includes

#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/framework/StdInInputSource.hpp>

//Pathan includes [the functions we use only use DOM - no Pathan internals]

#include <dom3-xpath/XPathEvaluator.hpp>
#include <dom3-xpath/XPathResult.hpp>
#include <dom3-xpath/XPathEvaluator.hpp>
#include <dom3-xpath/XPathNSResolver.hpp>

#include <context/Context.hpp>
#include <DOMutils/v2Utils.hpp>
#include <DOMutils/v2NSUtils.hpp>
#include <defs/NumUtils.hpp>

#include <DOMutils/XMLChBuilder.hpp>


//Local includes

#include "DOMTreeErrorReporter.hpp"
#include "DOMSerialise.hpp"


// functions
void welcome();
void usage();
int initDOM(bool doValidate, bool doNamespace);
void termDOM();
XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *loadXMLFile(char *filename);
int runExpression(XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *document, XPathEvaluator *evaluator, const XMLCh *expression, const string filename, bool escapeDoc);
XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *parseFromStdIn();
void printResultNodes(XPathResult *setSnapshot, const string filename);
void printResultPaths(XPathResult *setSnapshot, const string filename, const XPathNSResolver *nsResolver);
XMLCh *getPath(XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *node, const XPathNSResolver *context);

//global variables
XERCES_CPP_NAMESPACE_QUALIFIER XercesDOMParser *_xmlparser;
DOMTreeErrorReporter *_errHandler;
bool opt_p = false;

int main(int argc, char *argv[])
{
	bool escapeDoc = true;
	bool doValidate = true;
	bool doNamespace = true;

	// parse the command line arguments
	char *expression = NULL;
	int newArgC = 0;
	char **newArgV = (char **)malloc(argc * sizeof(char *));
	for(int i = 1; i < argc; ++i) {
		if(*argv[i] == '-') {
			// -t option, don't escape nodes
			if(argv[i][1] == 't' && argv[i][2] == '\0') {
				escapeDoc = false;
			}

			else if(argv[i][1] == 'v' && argv[i][2] == '\0') {
				doValidate = false;
			}

			else if(argv[i][1] == 'n' && argv[i][2] == '\0') {
				doNamespace = false;
			}

			// --path, -p option
			else if((argv[i][1] == 'p' && argv[i][2] == '\0') ||
					(argv[i][1] == '-' && argv[i][2] == 'p')) {
				opt_p = true;
			}
			// --help, -h option
			else if((argv[i][1] == 'h' && argv[i][2] == '\0') ||
					(argv[i][1] == '-' && argv[i][2] == 'h')) {
				welcome();
				usage();
				return 0;
			}
			else {
				welcome();
				usage();
				return 1;
			}
		}
		else if(expression == NULL) {
			expression = argv[i];
		}
		else {
			newArgV[newArgC++] = argv[i];
		}
	}

	// Check for bad command line arguments
	if(expression == NULL) {
		welcome();
		usage();
		return 1;
	}

	// initialisation
	if(initDOM(doValidate, doNamespace)) {
		return 1;
	}

        const XMLCh *expressionUnicode = XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(expression);

        XPathEvaluator *evaluator = NULL;
        try
        {


           //Initialise the Pathan interface

           //This is in it's own block since XPathEvaluator must destruct
           //before terminate dom is called, otherwise it will segfault
           {
              evaluator = XPathEvaluator::createEvaluator();

              // run the parsed expression on the document(s)
              if(newArgC == 0) {
                 XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *doc = parseFromStdIn();
                 if(runExpression(doc, evaluator, expressionUnicode, "stdin", escapeDoc) != 0) {
                    return 1;
                 }
              }
              else {
                 for(int i = 0; i < newArgC; ++i) {
                    // Load the XML file
                    XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *doc = loadXMLFile(newArgV[i]);
                    if(runExpression(doc, evaluator, expressionUnicode, newArgV[i], escapeDoc) != 0) {
                       return 1;
                    }
                 }
              }

           }
        }
        catch(XERCES_CPP_NAMESPACE_QUALIFIER DOMException& e)
        {
           delete evaluator;
           delete[] (XMLCh *)expressionUnicode;
           std::cerr << "XERCES_CPP_NAMESPACE_QUALIFIER DOMException, code " << e.code << ": " << XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(e.msg) << endl;
           return 1;
        }
        catch(...)
        {
           delete evaluator;
           delete[] (XMLCh *)expressionUnicode;
           std::cerr << "Unknown exception" << endl;
           return 1;
        }

        delete evaluator;
        delete[] (XMLCh *)expressionUnicode;

	// clean up and exit
	termDOM();

	free(newArgV);

	return 0;
}

int runExpression(XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *document, XPathEvaluator *evaluator, const XMLCh *expression, const string filename, bool escapeDoc)
{
	if(document!=NULL) {

		//Evaluate expression on document

		//Set trigger to document element, to pick up namespace bindings
		XPathNSResolver *nsResolver = evaluator->createNSResolver(document->getDocumentElement());
                XPathExpression *parsedExpression = NULL;
		XPathResult *result = NULL;

		try {
			//Parse expression
			parsedExpression = evaluator->createExpression(expression, nsResolver);

			//Evaluate expression
			//I'd rather have this in its own seperate try block by that has problems with the scope of parsedExpression
			result = parsedExpression->evaluate(document, XPathResult::ORDERED_NODE_SNAPSHOT_TYPE, result);

		}
		catch(const XPathException &e) {
                   delete nsResolver;
                   delete parsedExpression;
                   delete result;
                   std::cerr << e.getString() << endl;
                   return 1;
		}
        catch(...)
            {
                delete nsResolver;
                delete parsedExpression;
                throw;
            }
        delete nsResolver;
        delete parsedExpression;
        
        try
            {
                // print out the nodes obtained
                if(escapeDoc == true) {
                    XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *tmpNode = document;
                    v2Utils::escapeNode(tmpNode);
                }//if
                
                if(opt_p) {
                    printResultPaths(result, filename, nsResolver);
                }
                else {
                    printResultNodes(result, filename);
                }
            }
        catch(...)
            {
                delete result;
                throw;
                }
                delete result;
	}
	else {
		std::cerr << "Cannot open file " << filename << ".\n";
	}

	return 0;
}

void printResultNodes(XPathResult *result, const string filename)
{
	for(unsigned int i=0; i<result->getSnapshotLength(); i++) {
		XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *node = result->snapshotItem(i);

		std::cout << "<?xgrep-match [" << filename;
		std::cout << "]--------------------------------?>\n";
		std::cout << node << endl;
	}
}

void printResultPaths(XPathResult *result, const string filename, const XPathNSResolver *nsResolver)
{
	for(unsigned int i=0; i<result->getSnapshotLength(); i++) {
		XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *node = result->snapshotItem(i);

    const XMLCh* path = getPath(node, nsResolver);
    std::cout << filename << ": " << path << endl;
    delete[] (XMLCh *)path;
	}
}

XMLCh *getPath(XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *node, const XPathNSResolver *context)
{
	XMLChBuilder result;

	// preceding path
	XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *parent = v2NSUtils::getParent(node);
	if((parent!=NULL) && parent->getNodeType() != XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::DOCUMENT_NODE) {
           result.appendAndDelete(getPath(parent, context));
	}

	// slash then name
	result.appendLiteral("/");
	switch(node->getNodeType()) {
           case XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::TEXT_NODE: {
              result.appendLiteral("text()");
              break;
           }
           case XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::COMMENT_NODE: {
              result.appendLiteral("comment()");
              break;
           }
           case XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::ATTRIBUTE_NODE: {
              result.appendLiteral("@");
              // break left out
           }
           case XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::ELEMENT_NODE: {
              result.appendAndDelete(v2NSUtils::qualifyName(node->getPrefix(),
                                                            node->getLocalName()));
              break;
           }
           default: {
              result.appendLiteral("unknown node");
           }
	}

	// index
	if(node->getNodeType() != XERCES_CPP_NAMESPACE_QUALIFIER DOMNode::ATTRIBUTE_NODE) {
           result.appendLiteral("[");
           result.appendAndDelete(NumUtils::intAsString(v2Utils::indexNode(node)));
           result.appendLiteral("]");
	}

	return result.fetch();
}

// Set up xerces and the xml parser
int initDOM(bool doValidate, bool doNamespace)
{
	try{
		XERCES_CPP_NAMESPACE_QUALIFIER XMLPlatformUtils::Initialize();
	}
	catch (const XERCES_CPP_NAMESPACE_QUALIFIER XMLException& eXerces){
		char *pMsg = XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(eXerces.getMessage());
		std::cerr << "Error during Xerces-C initialisation.\n"
			<< "Xerces exception message: "
			<< pMsg << endl;
		delete pMsg;
		return 1;
	}

	_xmlparser = new XERCES_CPP_NAMESPACE_QUALIFIER XercesDOMParser();

	if(doValidate == true) {
		_xmlparser->setValidationScheme(XERCES_CPP_NAMESPACE_QUALIFIER XercesDOMParser::Val_Always);
	} else {
		_xmlparser->setValidationScheme(XERCES_CPP_NAMESPACE_QUALIFIER XercesDOMParser::Val_Never);
	}//if

	_xmlparser->setDoNamespaces(doNamespace);
	_xmlparser->setCreateEntityReferenceNodes(false);

	_errHandler = new DOMTreeErrorReporter();
	_xmlparser->setErrorHandler(_errHandler);

	return 0;
}

// kill xerces and the xml parser
void termDOM()
{
        delete _xmlparser; //parser must be deleted before calling Terminate
	delete _errHandler;
	XERCES_CPP_NAMESPACE_QUALIFIER XMLPlatformUtils::Terminate();
}

// parse an XML file from stdin
XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *parseFromStdIn()
{
	try {
		_xmlparser->parse(XERCES_CPP_NAMESPACE_QUALIFIER StdInInputSource());
	}
	catch(const XERCES_CPP_NAMESPACE_QUALIFIER XMLException& eParse) {
		std::cerr << "Error during script parsing from stdin"
			<< ".\nMessage: "
			<< eParse.getMessage()
			<< endl;

		return NULL;
	}
	catch(const XERCES_CPP_NAMESPACE_QUALIFIER SAXParseException& eParse) {
		std::cerr << "Error during script parsing from stdin"
			<< " at line " << eParse.getLineNumber()
			<< ", column " << eParse.getColumnNumber()
			<< ".\nMessage: "
			<< eParse.getMessage()
			<< endl;

		return NULL;
	}

	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *result = _xmlparser->getDocument();
	_xmlparser->reset();

	return result;
}

// parse an XML file to memory
XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *loadXMLFile(char *filename)
{
	try {
		_xmlparser->parse(filename);
	}
	catch(const XERCES_CPP_NAMESPACE_QUALIFIER XMLException& eParse) {
		std::cerr << "Error during script parsing, in file "
			<< filename << ".\nMessage: "
			<< eParse.getMessage()
			<< endl;

		return NULL;
	}
	catch(const XERCES_CPP_NAMESPACE_QUALIFIER SAXParseException& eParse) {
		std::cerr << "Error during script parsing, in file "
			<< filename << " at line " << eParse.getLineNumber()
			<< ", column " << eParse.getColumnNumber()
			<< ".\nMessage: "
			<< eParse.getMessage()
			<< endl;

		return NULL;
	}

	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *result = _xmlparser->getDocument();
	_xmlparser->reset();

	return result;
}

// print the welcome message
void welcome()
{
	std::cerr << "X-Grep 1.0 Beta." << endl;
	std::cerr << "(c) 2001-2003 DecisionSoft Limited.\n";
	std::cerr << "User documentation and licence at http://software.decisionsoft.com\n\n";
}

// print the usage message
void usage()
{
	std::cerr << "\nUsage: xgrep [-h] [-p] [-t] [-v] [-n] \"XPath Expression\" data1.xml [data2.xml] ..." << endl << endl;
	std::cerr << "-h : Show this display" << endl;
	std::cerr << "-p : Display an absolute xpath expression to query match" << endl;
	std::cerr << "-t : Don't escape contents of output nodes" << endl;
	std::cerr << "-v : Disable validation" << endl;
	std::cerr << "-n : Disable namespaces" << endl;
}

